""" Class to plot the results of the zfinder """

import csv
import warnings
from itertools import zip_longest

import numpy as np
import matplotlib.pyplot as plt
from astropy.wcs import WCS
from astropy.io import fits

from zfinder.template import calc_template_params, gaussf, find_lines
from zfinder.fft import calc_fft_params, double_damped_sinusoid
from zfinder.uncertainty import z_uncert
from zfinder.utils import wcs2pix, generate_square_pix_coords, longest_decimal

warnings.filterwarnings(action="ignore", message='Some errors were detected !')
warnings.filterwarnings("ignore", module='astropy.wcs.wcs')

class Plotter():
    """ 
    Class to plot the results of zfinder. Includes methods of plotting exported csv data
    
    Parameters
    ----------
    showfig : bool, optional
        Whether to show the figure after plotting. Default is True.
    
    savefig : bool, optional
        Whether to save the figure after plotting. Default is True.
    
    Attributes
    ----------
    unit_prefixes : dict
        Dictionary of unit prefixes for plotting
    
    Methods
    -------
    calc_best_z()
        Calculate the best redshift
    
    plot_chi2()
        Plot the chi-squared vs redshift
    
    plot_template_flux()
        Plot the template flux
    
    plot_fft_flux()
        Plot the fft flux
        
    plot_heatmap()
        Plot a heatmap of the redshifts
    
    export_template_data()
        Export the Template fit data to a csv file
    
    export_fft_data()
        Export the FFT data to a csv file 
        
    plot_chi2_fromcsv()
        Plot the chi-squared vs redshift from a csv file
    
    plot_template_flux_fromcsv()
        Plot the template flux from a csv file
    
    plot_fft_flux_fromcsv()
        Plot the fft flux from a csv file
    
    plot_heatmap_fromcsv()
        Plot a heatmap of the redshifts from a csv file
    
    Examples
    --------
    >>> # After csv files exported with zfinder
    >>> source = Plotter()
    >>> source.plot_chi2_fromcsv('template.csv')
    >>> source.plot_template_flux_fromcsv()
    >>> 
    >>> source.plot_chi2_fromcsv('fft.csv')
    >>> source.plot_fft_flux_fromcsv()
    >>> 
    >>> source.plot_heatmap_fromcsv('template_per_pixel.csv')
    >>> source.plot_heatmap_fromcsv('fft_per_pixel.csv')
    """
    
    unit_prefixes = {
        -24 : 'y', 
        -21 : 'z',
        -18 : 'a',
        -15 : 'f',
        -12 : 'p',
        -9 : 'n',
        -6 : '\u03BC',
        -3 : 'm',
        0 : '', 
        3 : 'k',
        6 : 'M',
        9 : 'G',
        12 : 'T',
        15 : 'P',
        18 : 'E', 
        21 : 'Z',
        24 : 'Y'}
    
    def __init__(self, showfig=True, savefig=False):
        self._showfig = showfig
        self._savefig = savefig
        self._template_best_z = None
        self._fft_best_z = None
        self._round_to = 2
        self._freq_exp = None
        self._flux_exp = None
    
    def calc_best_z(self, z, chi2, title=None):
        """ 
        Calculate the best redshift
        
        Parameters
        ----------
        z : list
            Array of redshifts
        
        chi2 : list
            Array of chi-squared values
        
        Returns
        -------
        best_z : float
            The best redshift
        """
        best_z = z[np.argmin(chi2)]
        if title == 'Template':
            self._template_best_z = best_z
        elif title.upper() == 'FFT':
            self._fft_best_z = best_z
        return best_z
            
    def plot_chi2(self, z, dz, chi2, title):
        """ Plot the chi-sqaured vs redshift """
        self._z = z
        self._chi2 = chi2
        min_chi2 = min(chi2)
        best_z = self.calc_best_z(z, chi2, title)
        self._round_to = len(str(dz).split('.')[1])
        plt.figure(figsize=(15,7))
        plt.plot(z, chi2, color='black', label='$\chi^2_r$')
        plt.plot(best_z, min_chi2, 'bo', markersize=5, label='Best Fit')
        plt.title(f'{title} $\chi^2_r$ = {round(min_chi2, 2)} @ z={round(best_z, self._round_to)}', fontsize=15)
        plt.xlabel('Redshift', fontsize=15)
        plt.ylabel('$\chi^2_r$', x=0.01, fontsize=15)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        plt.yscale('log')
        plt.legend()
        if self._savefig:
            plt.savefig(f'{title.lower()}_chi2.png')
        if self._showfig:
            plt.show()
    
    @staticmethod
    def _plot_sslf_lines(frequency, flux):
        """ Helper function to plot sslf lines in the found flux """
        peaks, snrs, scales = find_lines(flux)
        text_offset_high = max(flux)/20
        text_offset_low = 0.4*text_offset_high
        for i, line in enumerate(peaks):
            x = frequency[line]
            y = flux[line]
            plot, = plt.plot(x, y, 'bo')
            plt.text(x, y+text_offset_high, f'snr={snrs[i]}', color='blue')
            plt.text(x, y+text_offset_low, f'scale={scales[i]}', color='blue')
            if i == 0:
                plot.set_label('Lines')
        
    def plot_template_flux(self, transition, frequency, freq_exp, flux, flux_exp):
        """ 
        Plot the template flux 
        
        Parameters
        ----------
        transition : float
            The transition frequency
        
        frequency : list
            Array of frequencies
        
        freq_exp : int
            The exponent of the frequency 
        
        flux : list
            Array of fluxes
        
        flux_exp : int
            The exponent of the flux
        """
        self._frequency = frequency
        self._flux = flux
        self._freq_exp = freq_exp
        self._flux_exp = flux_exp
        plt.figure(figsize=(15,7))
        plt.plot(frequency, np.zeros(len(frequency)), color='black', linestyle=(0, (5, 5)))
        plt.plot(frequency, flux, color='black', drawstyle='steps-mid')
        if self._template_best_z is None:
            raise ValueError("No best redshift found. Run Plotter.plot_chi2() first.")
        x0 = transition/(1+self._template_best_z)
        self._params, covars = calc_template_params(frequency, flux, x0)
        self._p_err = np.sqrt(np.diag(covars)) # calculate the error on the parameters
        plt.plot(frequency, gaussf(frequency, *self._params, x0), color='red', label='Template Fit')
        self._plot_sslf_lines(frequency, flux)
        plt.margins(x=0)
        plt.fill_between(frequency, flux, 0, where=(np.array(flux) > 0), color='gold', alpha=0.75, label='Aperture Flux')
        plt.title(f'Template Fit z={round(self._template_best_z, self._round_to)}', fontsize=15)
        plt.xlabel(f'Frequency $({self.unit_prefixes[freq_exp]}Hz)$', fontsize=15)
        plt.ylabel(f'Flux $({self.unit_prefixes[flux_exp]}Jy)$', fontsize=15)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        plt.legend()
        if self._savefig:
            plt.savefig('template_flux.png')
        if self._showfig:
            plt.show()
        
    def plot_fft_flux(self, transition, frequency, ffreq, fflux):
        """ 
        Plot the fft flux
        
        Parameters
        ----------
        transition : float
            The transition frequency
        
        frequency : list
            Array of frequencies
        
        ffreq : list
            Array of fft frequencies
        
        fflux : list
            Array of fft fluxes
        """
        self._ffreq = ffreq
        self._fflux = fflux
        plt.figure(figsize=(15,7))
        plt.plot(ffreq, fflux, color='black', drawstyle='steps-mid', label='FFT Flux')
        plt.plot(ffreq, np.zeros(len(fflux)), color='black', linestyle=(0, (5, 5)))
        if self._fft_best_z is None:
            raise ValueError("No best redshift found. Run Plotter.plot_chi2() first.")
        self._params, covars = calc_fft_params(transition, ffreq, fflux, self._fft_best_z, frequency[0])
        self._p_err = np.sqrt(np.diag(covars))
        plt.plot(ffreq, double_damped_sinusoid(ffreq, *self._params, 
            self._fft_best_z, frequency[0], transition), color='red', label='FFT Fit')
        plt.margins(x=0)
        plt.title(f'FFT Fit z={round(self._fft_best_z, self._round_to)}', fontsize=15)
        plt.xlabel('Scale', fontsize=15)
        plt.ylabel('Amplitude', fontsize=15)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        plt.legend()
        if self._savefig:
            plt.savefig('fft_flux.png')
        if self._showfig:
            plt.show()
            
    @staticmethod
    def _export_heatmap_data(filename, mode, data):
        """ Export the redshifts, velocities, and fluxes to a csv file """
        with open(filename, mode, newline='') as f:
            wr = csv.writer(f)
            wr.writerows(data)
            wr.writerow('')
    
    def plot_heatmap(self, ra, dec, hdr, data, size, z, title, aperture_radius, flux_limit, export=False, subsize=None, contfile=None):
        """ 
        Plot a heatmap of the redshifts
        
        Parameters
        ----------
        ra : list[string]
            Target right ascension
        
        dec : list[string]
            Target declination
        
        hdr : astropy.io.fits.header.Header
            The header of the fits file
        
        data : astropy.io.fits.hdu.image.PrimaryHDU
            The data of the fits file
        
        size : int
            The size of the square in pixels
            
        z : list
            Array of redshifts
        
        title : string
            The title method of the plot. Either 'Template' or 'FFT'
        
        aperture_radius : float
            The radius of the aperture in pixels
        
        flux_limit : float
            The flux limit to mask the velocities
        
        export : bool, optional
            Whether to export the redshifts to a csv file. Default is True.  
        
        subsize : int, optional
            Plot a smaller heatmap of the redshifts. Default is None.
        """
        # Calculate the velocities
        target_z = np.take(z, z.size // 2) # redshift of the target ra and dec
        velocities = 3*10**5*((((1 + target_z)**2 - 1) / ((1 + target_z)**2 + 1)) - (((1 + z)**2 - 1) / ((1 + z)**2 + 1))) # km/s
        
        # Need to get x and y coordinates to plot the heatmap with bounds for correct ra and dec
        target_pix_ra_dec = wcs2pix(ra, dec, hdr)
        x, y = generate_square_pix_coords(size, *target_pix_ra_dec, aperture_radius)
        
        # Mask velocities lower than the flux limit
        if contfile is not None:
            data_summed = fits.getdata(contfile)
        else:
            data_summed = np.sum(np.maximum(data, 0), axis=0)
        uy = np.round(np.unique(y)).astype(int)
        ux = np.round(np.unique(x)).astype(int)
        fluxes = data_summed[uy][:, ux]
        mask = fluxes < flux_limit
        if subsize is not None:
            mask = self._extract_centered_subarray(mask, subsize)
            uy = self._extract_centered_subarray_1d(uy, subsize)
            ux = self._extract_centered_subarray_1d(ux, subsize)
        velocities[mask] = np.nan
        
        if export:
            self._export_heatmap_data(f'{title.lower()}_per_pixel.csv', 'a', velocities) # export redshifts to csv
            self._export_heatmap_data(f'{title.lower()}_per_pixel.csv', 'a', fluxes) # export redshifts to csv

        cmap = plt.cm.seismic
        cmap.set_bad('black')
        scale_velo = np.nanmax(np.abs(velocities))
        
        # velocities = np.flipud(velocities)
        w = WCS(hdr, naxis=2)
        plt.figure(figsize=(7,5))
        plt.subplot(projection=w)
        hm = plt.imshow(np.flipud(velocities), cmap=cmap, interpolation='nearest', vmin=-scale_velo, vmax=scale_velo,
                extent=[ux[0], ux[-1], uy[0], uy[-1]], origin='lower')
        cbar = plt.colorbar(hm)
        cbar.ax.set_ylabel('km/s', fontsize=15)
        cbar.ax.tick_params(labelsize=15)
        plt.title(f'{title} Per Pixel', fontsize=15)
        plt.xlabel('RA', fontsize=15)      
        plt.ylabel('DEC', fontsize=15)   
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15) 
        if self._savefig:
            plt.savefig(f'{title.lower()}_per_pixel.png')
        if self._showfig:
            plt.show()
    
    def plot_coords(self, x_centre, y_centre, x_coords, y_coords, radius, fitsfile=None):
        if fitsfile is not None: 
            hdr = fits.getheader(fitsfile)
            w = WCS(hdr, naxis=2)
            fig, ax = plt.subplots(subplot_kw={'projection': w})
        else:
            fig, ax = plt.subplots()
        circ = plt.Circle((x_centre, y_centre), radius, fill=False, color='blue', label='_nolegend_')
        fig.set_figwidth(7)
        fig.set_figheight(7)
        plt.subplots_adjust(left=0.2, right=0.8, top=0.8, bottom=0.2)
        ax.add_patch(circ)
        plt.scatter(x_coords, y_coords, color='blue', label='Random')
        plt.scatter(x_centre, y_centre, color='black', label='Target')
        plt.title(f'{len(x_coords)} random points')
        plt.xlim(-radius-1+x_centre, radius+1+x_centre)
        plt.ylim(-radius-1+y_centre, radius+1+y_centre)
        plt.xlabel('RA', fontsize=15)
        plt.ylabel('DEC', fontsize=15)
        plt.legend(loc='upper left')
        if self._savefig:
            plt.savefig('Point Distribution.png', dpi=200)
        if self._showfig:
            plt.show()

    @staticmethod
    def _write_method_to_csv(filename, headings, data):
        with open(filename, 'w', newline='') as f:
            wr = csv.writer(f)
            wr.writerow(headings)
            wr.writerows(data)       
        
    def _z_uncert(self, sigma, flux):
        """ Caclulate the uncertainty on the best fitting redshift """
        peaks, _, _ = find_lines(flux)
        reduced_sigma = sigma**2 / (len(flux) - 2*len(peaks) - 1)
        neg, pos = z_uncert(self._z, self._chi2, reduced_sigma)
        return neg, pos
            
    def _calculate_results(self, sigma, best_z, flux):
        z_low_err, z_up_err = self._z_uncert(sigma, flux)
        results = [[z_low_err], [round(best_z, self._round_to)], 
                [z_up_err], [self._params[0]], [self._p_err[0]], [self._params[1]], [self._p_err[1]]]
        if self._freq_exp is None:
            return results, None
        exponents = [[self._freq_exp], [self._flux_exp]]   
        return results, exponents

    def _export_template_data(self, fitsfile, ra, dec, aperture_radius, transition, filename='template.csv', sigma=1, flux_uncert=1):
        """ 
        Export the Template fit data to a csv file 
        
        Ensure that Plotter.plot_template_flux() and Plotter.plot_chi2() have been run first.
        
        Parameters
        ----------
        filename : string, optional
            The filename of the csv file. Default is 'template.csv'.
        
        sigma : float, optional
            The sigma value for the uncertainty on the redshift. Default is 1.
        
        flux_uncert : float, optional
            The uncertainty on the flux. Default is 1.
        """
        results, exponents = self._calculate_results(sigma, self._template_best_z, self._flux)
        headings = ['z_low_err', 'z', 'z_up_err', 'amp', 'amp_err', 'std_dev', 
                    'std_dev_err', 'fitsfile', 'ra', 'dec', 'aperture_radius', 'transition', 
                    'dz', 'chi2_r', 'freq', 'flux', 'flux_uncert', 'freq_exp', 'flux_exp']
        common = [[fitsfile], [ra], [dec], [aperture_radius], [transition]]
        data = [*zip_longest(*results, *common, self._z, self._chi2, self._frequency, self._flux, flux_uncert, *exponents, fillvalue='')]
        self._write_method_to_csv(filename, headings, data)

    def _export_fft_data(self, fitsfile, ra, dec, aperture_radius, transition, filename='fft.csv', sigma=1, frequency=None, flux=None, flux_uncert=1):
        """ 
        Export the FFT data to a csv file 
        
        Ensure that Plotter.plot_fft_flux() and Plotter.plot_chi2() have been run first.
        
        Parameters
        ----------
        filename : string, optional
            The filename of the csv file. Default is 'fft.csv'.
        
        sigma : float, optional
            The sigma value for the uncertainty on the redshift. Default is 1.
        
        flux_uncert : float, optional
        """
        results, _ = self._calculate_results(sigma, self._fft_best_z, flux)
        headings = ['z_low_err', 'z', 'z_up_err', 'amp', 'amp_err', 'std_dev',
                    'std_dev_err', 'fitsfile', 'ra', 'dec', 'aperture_radius', 'transition',
                    'dz', 'chi2_r', 'frequency', 'flux', 'ffreq', 'fflux', 'fflux_uncert']
        if type(flux_uncert) == int:
            flux_uncert = [flux_uncert]
        common = [[fitsfile], [ra], [dec], [aperture_radius], [transition]]
        data = [*zip_longest(*results, *common, self._z, self._chi2, frequency, flux, self._ffreq, self._fflux, flux_uncert, fillvalue='')]
        self._write_method_to_csv(filename, headings, data)
    
    def plot_chi2_fromcsv(self, filename):
        """ Plot the chi-squared vs redshift from a csv file """
        z, chi2 = np.genfromtxt(filename, delimiter=',', skip_header=1, usecols=(12,13)).T
        z = z[~np.isnan(z)]
        chi2 = chi2[~np.isnan(chi2)]
        dz = longest_decimal(z)
        self.plot_chi2(z, dz, chi2, filename.split('.')[0].capitalize())
    
    def plot_template_flux_fromcsv(self, filename='template.csv'):
        """ Plot the template flux from a csv file """
        transition, frequency, flux, freq_exp, flux_exp = np.genfromtxt(filename, delimiter=',', skip_header=1, usecols=(11, 14, 15, 17, 18)).T
        transition = transition[0]
        frequency = frequency[~np.isnan(frequency)]
        flux = flux[~np.isnan(flux)]
        freq_exp = int(freq_exp[0])
        flux_exp = int(flux_exp[0])
        self.plot_template_flux(transition, frequency, freq_exp, flux, flux_exp)
    
    def plot_fft_flux_fromcsv(self, filename='fft.csv'):
        """ Plot the fft flux from a csv file """
        transition, frequency, ffreq, fflux = np.genfromtxt(filename, delimiter=',', skip_header=1, usecols=(11, 14, 16, 17)).T
        transition = transition[0]
        frequency = frequency[~np.isnan(frequency)]
        ffreq = ffreq[~np.isnan(ffreq)]
        fflux = fflux[~np.isnan(fflux)]
        self.plot_fft_flux(transition, frequency, ffreq, fflux)
    
    def plot_heatmap_fromcsv(self, filename, subsize=None, flux_limit=None, contfile=None):
        """ Plot a heatmap of the redshifts from a csv file """
        dtypes = [('fitsfile', 'U100'), ('ra', 'U100'), ('dec', 'U100'), ('aperture_radius', 'f8'),
          ('transition', 'f8'), ('size', 'i4'), ('flux_limit', 'f8')]
        fitsfile, ra, dec, aperture_radius, _, size, flux_lim = \
            np.genfromtxt(filename, delimiter=',', skip_header=1, dtype=dtypes, max_rows=2, invalid_raise=False).tolist()
        z = np.genfromtxt(filename, delimiter=',', skip_header=3, skip_footer=size*2)
        if flux_limit is not None:
            flux_lim = flux_limit
        if subsize is not None:
            z = self._extract_centered_subarray(z, subsize)
        hdr = fits.getheader(fitsfile)
        data = fits.getdata(fitsfile)[0]
        self.plot_heatmap(ra, dec, hdr, data, size, z, filename.split('_')[0].capitalize(), aperture_radius, flux_lim, subsize=subsize, contfile=contfile)
    
    def plot_coords_fromcsv(self, filename='fft_uncertainty.csv'):
        """ Plot the distribution of random points from a csv file """
        x_centre, y_centre, x_coords, y_coords, radius = np.genfromtxt(filename, delimiter=',', skip_header=1, usecols=(1,2,3,4,8)).T
        fitsfile = np.genfromtxt(filename, delimiter=',', skip_header=1, dtype='U100', usecols=(10)).T
        x_centre = x_centre[0]
        y_centre = y_centre[0]
        x_coords = x_coords[~np.isnan(x_coords)]
        y_coords = y_coords[~np.isnan(y_coords)]
        radius = radius[0]
        fitsfile = fitsfile[0]
        self.plot_coords(x_centre, y_centre, x_coords, y_coords, radius, fitsfile)
    
    @staticmethod
    def _extract_centered_subarray(original_array, new_size):
        """ Extract a centered subarray from a 2D array """
        start_row = (original_array.shape[0] - new_size) // 2
        end_row = start_row + new_size
        start_col = (original_array.shape[1] - new_size) // 2
        end_col = start_col + new_size
        centered_subarray = original_array[start_row:end_row, start_col:end_col]
        return centered_subarray
    
    @staticmethod
    def _extract_centered_subarray_1d(original_array, new_size):
        """ Extract a centered subarray from a 1D array """
        start_index = (original_array.shape[0] - new_size) // 2
        end_index = start_index + new_size
        centered_subarray = original_array[start_index:end_index]
        return centered_subarray


"""
Module for finding the redshift of a source using the gaussian template shifting method.
"""

import warnings
from multiprocessing import Pool

import numpy as np
from tqdm import tqdm
from sslf.sslf import Spectrum
from scipy.signal import find_peaks
from scipy.optimize import curve_fit

warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide", category=RuntimeWarning)
warnings.filterwarnings("ignore", message="invalid value encountered in divide", category=RuntimeWarning)

def gaussf(x, a, s, x0):
    """ Function to fit a sum of gaussians """
    y = 0
    for i in range(1,30):
        y += (a * np.exp(-((x-i*x0) / s)**2)) # i = 1,2,3 ... 29,30
    return y

def _calc_reduced_chi2(flux, flux_expected, flux_uncertainty, gauss_lines):
    """ Calculate the reduced chi-squared: chi2_r = chi2 / (num_points - free_params - 1) """
    gauss_dofs = 2*len(gauss_lines) # 2 DoF per gaussian (amp, std)
    DoF = len(flux) - gauss_dofs - 1 # Calculate the degrees of freedom

    # Calculate chi-squared
    chi2 = sum(((flux - flux_expected) / flux_uncertainty)**2)

    # Calculate reduced chi-squared
    reduced_chi2 = chi2 / DoF
    return reduced_chi2

def _penalise_chi2(gauss_lines, sslf_lines):
    """ Penalise chi-squared values that do not fit to lines found with sslf line finder """
    penalising_factor = 1.2 # factor multiplying bad fits
    non_penalising_factor = 1 # factor multiplying good fits
    tolerance = 3 # minimum number of channels separating gauss and sslf lines
    
    # Sort Ascending
    gauss_lines = np.sort(gauss_lines)
    sslf_lines = np.sort(sslf_lines)

    # If the number of lines are not equal, return penalising factor
    if len(sslf_lines) != len(gauss_lines):
        return penalising_factor

    # If the lines are not within a range of each other, return penalising factor
    for bline, gline in zip(sslf_lines, gauss_lines):
        if gline not in range(bline-tolerance, bline+tolerance):
            return penalising_factor
        
    # If lines are within range, return non penalising factor
    else:
        return non_penalising_factor

def find_lines(flux):
    """ Create a line finder to find significant points """
    s = Spectrum(flux)
    s.find_cwt_peaks(scales=np.arange(4,10), snr=3)
    spec_peaks = s.channel_peaks

    # Calculate the ratio of the snrs and scales
    snrs = [round(i,2) for i in s.peak_snrs] # the snrs of the peaks
    scales = [i[1]-i[0] for i in s.channel_edges] # the scales of the peaks

    # Sort lists
    sorted_lists = zip(*sorted(zip(spec_peaks, snrs, scales)))
    try:
        spec_peaks, snrs, scales = map(list, sorted_lists)
    except ValueError:
        spec_peaks, snrs, scales = [], [], [] # when there are 0 lines found
    return spec_peaks, snrs, scales

def calc_template_params(frequency, flux, observed_transition):
    params, covars = curve_fit(lambda x, a, s: gaussf(x, a, s, x0=observed_transition), 
        frequency, flux, bounds=[[0, (1/8)], [2*max(flux), (2/3)]], absolute_sigma=True)
    return params, covars    
    
def _process_template_chi2_calculations(transition, frequency, flux, flux_uncertainty, sslf_lines, dz):
    """ Use multiprocessing to significantly speed up chi2 calculations """
    # Calculate the frequency of the first observed transition line
    observed_transition = transition / (1 + dz)

    # Fit a gaussian template to the flux
    try:
        params, _ = calc_template_params(frequency, flux, observed_transition) # best fit
    except RuntimeError:
        chi2 = sum((flux)**2) * 1.2
        return chi2
    
    # Calculate the flux of a perfect function fit
    flux_expected = gaussf(frequency, a=params[0], s=params[1], x0=observed_transition)

    # Finds the list of Gaussians lines present at this redshift and inside the window (in channel no.#)
    gauss_lines = find_peaks(flux_expected)[0] # E.g. = [60, 270]

    # Calculate reduced chi2
    multiplier = _penalise_chi2(gauss_lines, sslf_lines)
    reduced_chi2 = _calc_reduced_chi2(flux, flux_expected, flux_uncertainty, gauss_lines) * multiplier
    return reduced_chi2

def _mp_template_zfind(transition, frequency, flux, flux_uncertainty, z, sslf_lines, verbose=True):
    """ Process the Template chi2 calculations in parallel """
    with Pool() as pool:
        jobs = [pool.apply_async(_process_template_chi2_calculations, 
            (transition, frequency, flux, flux_uncertainty, sslf_lines, dz)) for dz in z]
        chi2 = [result.get() for result in tqdm(jobs, disable=not verbose)]
    return chi2

def _serial_template_zfind(transition, frequency, flux, flux_uncertainty, z, sslf_lines, verbose=True):
    """ Process the Template chi2 calculations serially """
    chi2 = [_process_template_chi2_calculations(transition, frequency, flux, flux_uncertainty, sslf_lines, dz) for dz in tqdm(z, disable=not verbose)]
    return chi2

def template_zfind(transition, frequency, flux, flux_uncertainty=1, z_start=0, dz=0.01, z_end=10, verbose=True, parallel=True):
    """
    Using the gaussian template shifting method, calculate the reduced chi-squared at every change
    in redshift.

    Parameters
    ----------
    z_start : float, optional
        The redshift to start calculated chi2 fits at. Default=0

    dz : float, optional
        Change in redshift to iterature through. Default=0.01

    z_end : float, optional
        Final redshift to start calculating chi2 at. Default=10

    verbose : bool, optional
        If True, print progress. Default=True

    Returns
    -------
    z : list
        List of redshift values that were iterated through

    chi2 : list
        List of calculated chi-squared values.

    Example
    -------
    >>> z, chi2 = template_zfind()
    >>> lowest_index = np.argmin(chi2)
    >>> best_fit_redshift = z[lowest_index]
    """
    # Create a list of redshifts to iterate through
    z = np.arange(z_start, z_end+dz, dz)
    sslf_lines, _, _ = find_lines(flux) # E.g. = [60, 270]

    # Parallelise slow loop to execute much faster
    if verbose:
        print('Calculating Template fit chi-squared values...')
    if parallel:
        chi2 = _mp_template_zfind(transition, frequency, flux, flux_uncertainty, z, sslf_lines, verbose=verbose)
    else:
        chi2 = _serial_template_zfind(transition, frequency, flux, flux_uncertainty, z, sslf_lines, verbose=verbose)
    return z, chi2

def template_per_pixel(transition, frequency, all_flux, all_flux_uncertainty, z_start=0, dz=0.01, z_end=10, size=3):
    """ Perform the Template fitting method on all flux values """
    print('Calculating all Template fit chi-squared values...')
    verbose, parallel = False, False
    with Pool() as pool:
        jobs = [pool.apply_async(template_zfind, (transition, frequency, flux, flux_uncert, z_start, dz, z_end, verbose, parallel)) 
            for flux, flux_uncert in zip(all_flux, all_flux_uncertainty)]
        results = [res.get() for res in tqdm(jobs)]
    all_z = [z[np.argmin(chi2)] for z, chi2 in results]
    z = np.reshape(all_z, (size, size))
    return z


"""
Utility functions for zfinder
"""

import numpy as np

from astropy.io import fits
from astropy.wcs import WCS
from astropy import units as u
from astropy.coordinates import SkyCoord, Angle
from scipy.spatial.distance import cdist

def wcs2pix(ra, dec, hdr):
    """ Convert RA, DEC to x, y pixel coordinates """
    # Get the RA & DEC in degrees
    c = SkyCoord(ra, dec, unit=(u.hourangle, u.degree))
    ra = Angle(c.ra).degree
    dec = Angle(c.dec).degree

    # Convert RA & DEC to pixel world coordinates
    wcs = WCS(hdr, naxis=2)
    x, y = wcs.all_world2pix(ra, dec, 1)
    return [x, y]

def get_sci_exponent(number):
    """ Find the scientific exponent of a number """
    abs_num = np.abs(number)
    base = np.log10(abs_num)  # Log rules to find exponent
    exponent = int(np.floor(base))  # convert to floor integer
    return exponent

def get_eng_exponent(number):
    """ 
    Find the nearest power of 3 (lower). In engineering format,
    exponents are multiples of 3.
    """
    exponent = get_sci_exponent(number)  # Get scientific exponent
    for i in range(3):
        if exponent > 0:
            unit = exponent-i
        else:
            unit = exponent+i
        if unit % 3 == 0:  # If multiple of 3, return it
            return unit

def average_zeroes(array):
    """ Average zeroes with left & right values in a list """
    for i, val in enumerate(array):
        if val == 0:
            try:
                array[i] = (array[i-1] + array[i+1])/2
            except IndexError:
                array[i] = (array[i-2] + array[i-1])/2
    return array

def radec2str(ra, dec):
    """ Convert RA and DEC to a string """
    ra = Angle(ra, unit=u.degree).to_string(unit=u.hour, sep=':', precision=4)
    dec = Angle(dec, unit=u.degree).to_string(unit=u.degree, sep=':', precision=4)
    return ra, dec

def generate_square_pix_coords(size, target_x, target_y, aperture_radius=0.5):
    """ Generate a list of coordinates for a square of pixel coordinates around a target x,y pixel """
    matrix = np.arange(size) - size//2
    matrix = matrix * aperture_radius * 2
    x, y = np.meshgrid(matrix + target_x, matrix + target_y)
    x, y = x.ravel(), y.ravel()
    return x, y

def generate_square_world_coords(fitsfile, ra, dec, size, aperture_radius):
    """ Generate a list of coordinates for a square of ra & dec around a target ra & dec """
    # Generate x, y pix coordinates around target ra & dec
    hdr = fits.getheader(fitsfile)
    target_pix_ra_dec = wcs2pix(ra, dec, hdr)
    x, y = generate_square_pix_coords(size, *target_pix_ra_dec, aperture_radius)

    # Convert x, y pix coordinates to world ra and dec
    wcs = WCS(hdr, naxis=2)
    ra, dec = wcs.all_pix2world(x, y, 1)
    ra, dec = radec2str(ra, dec)
    return ra, dec

def longest_decimal(numbers):
    """ Find the longest decimal place in a list of numbers """
    string_numbers = np.array2string(numbers)
    length_numbers = [len(i) for i in string_numbers.split()]
    return numbers[np.argmax(length_numbers)]

def gen_random_coords(n, radius, centre, min_spread):
    """ Generate a list of random x, y tuple coordinates inside a circle."""
    coords = [centre]
    
    # Iterate through the number of coords.
    for _ in range(n):
        
        # Keep generating the current point until it is at least the minimum distance away from all 
        while True:
            theta = 2 * np.pi * np.random.random() # choose a random direction
            r = radius * np.random.random() # choose a random radius

            # Convert coordinates to cartesian
            x = r * np.cos(theta) + centre[0]
            y = r * np.sin(theta) + centre[1]

            # Find the distance between all the placed coords
            distances = cdist([[x,y]], coords, 'euclidean')
            min_distance = min(distances[0])
            
            # If the minimum distance is satisfied for all coords, go to next point
            if min_distance >= min_spread:
                coords.append([x,y])
                break
    x, y = np.transpose(coords[1:])
    return x, y


"""
Module for finding the redshift of a source using the fourier transform method.
"""

from multiprocessing import Pool

import numpy as np
from tqdm import tqdm
from scipy.signal import find_peaks
from scipy.optimize import curve_fit

from zfinder.template import gaussf

def double_damped_sinusoid(x, a, s, z, nu, f):
    """ FFT Fitting function """
    A = 22.09797605*a*s
    C = 1/(np.pi*s)
    N = np.floor(((1+z)*nu/f)+1)
    p = 2*np.pi*(N*f/(1+z) - nu)
    q = 2*np.pi*((N+1)*f/(1+z) - nu)
    y = A*np.exp(-(x / C)**2) * (np.cos(p*x) + np.cos(q*x))
    return y

def calc_fft_params(transition, ffreq, fflux, dz, x0):
    """ Find the best fitting parameters for the FFT """
    params, covars = curve_fit(lambda x, a, s: double_damped_sinusoid(x, a, s, z=dz, 
        nu=x0, f=transition), ffreq, fflux, bounds=[[0, 0], [max(fflux), 2]])
    return params, covars

def fft(frequency, flux):
    """ 
    Performs the fast fourier transform on the frequency and flux axis 
    
    Parameters
    ----------
    frequency : list
        Frequency axis values found with fits2flux `get_freq()`
    
    flux : list
        Flux axis values found with fit2flux `get_flux()`
    
    Returns
    -------
    ffreq, fflux : list
        The Fourier Transformed frequency and flux axes respectively
    """
    N = 10*len(flux) # Number of sample points
    T = frequency[1]-frequency[0] # sample spacing

    # Fourier transformed data
    fflux = np.fft.rfft(flux, N).real
    ffreq = np.fft.rfftfreq(N, T)
    return ffreq, fflux

def _calc_all_num_gauss(transition, frequency, dz):
    """ Calculate the number of gaussians inside the window (x-axis) """
    loc = transition/(1+dz)
    f_exp = gaussf(frequency, a=0.5, s=0.5, x0=loc)

    # Caclulate the number of gaussians overlayed
    gauss_peaks = find_peaks(f_exp)[0]
    return gauss_peaks

def _process_fft_chi2_calculations(transition, frequency, uncertainty, ffreq, fflux, dz, x0):
    """ Use multiprocessing to significantly speed up chi2 calculations """
    # Find the best fitting parameters at this redshift. If fails return bad chi2
    try:
        params, _ = calc_fft_params(transition, ffreq, fflux, dz, x0)
    except RuntimeError:
        chi2 = sum((fflux)**2) * 1.2 # 1.2 is the penalising factor
        return chi2

    # Calulate chi-squared
    fflux_obs = double_damped_sinusoid(ffreq, a=params[0], s=params[1], z=dz, nu=x0, f=transition)

    # Find the number of gaussians that would be overlayed at this redshift
    gauss_peaks = _calc_all_num_gauss(transition, frequency, dz)

    chi2 = sum(((fflux - fflux_obs) / uncertainty)**2) # chi-squared
    reduced_chi2 = chi2 / (len(fflux_obs) - 2*len(gauss_peaks) - 1)
    return reduced_chi2

def _mp_fft_zfind(transition, frequency, uncertainty, ffreq, fflux, z, verbose=True):
    """ Process the FFT chi2 calculations in parallel """
    with Pool() as pool:
        jobs = [pool.apply_async(_process_fft_chi2_calculations, (transition, frequency, uncertainty, ffreq, fflux, dz, frequency[0])) for dz in z]
        chi2 = [res.get() for res in tqdm(jobs, disable=not verbose)]
    return chi2

def _serial_fft_zfind(transition, frequency, uncertainty, ffreq, fflux, z, verbose=True):
    """ Process the FFT chi2 calculations serially """
    chi2 = [_process_fft_chi2_calculations(transition, frequency, uncertainty, ffreq, fflux, dz, frequency[0]) for dz in tqdm(z, disable=not verbose)]
    return chi2

def fft_zfind(transition, frequency, flux, uncertainty=1, z_start=0, dz=0.01, z_end=10, verbose=True, parallel=True):
    """ 
    Finds the best redshift by performing the fast fourier transform on the flux data. The
    chi-squared is caclulated at every redshift by iterating through delta-z. The most 
    likely redshift corresponds to the minimum chi-squared.
    
    Parameters
    ----------        
    z_start : int, optional
        The first value in the redshift list. Default = 0
        
    dz : float, optional
        The change in redshift. Default = 0.01
    
    z_end : int, optional
        The final value in the redshift list. Default = 10
    
    verbose : bool, optional
        If True, print progress. Default = False
    
    Returns
    -------
    z : list
        The list of redshifts that was used to calculate the chi-squared
    
    chi2 : list
        A list of calculated chi-squared values
    """
    
    # Initialise lists
    z = np.arange(z_start, z_end+dz, dz)
      
    # Fourier transform frequency and flux
    ffreq, fflux = fft(frequency, flux)

    # Parallelise the chi2 calculations
    if verbose:
        print('Calculating FFT fit chi-squared values...')
    if parallel:
        chi2 = _mp_fft_zfind(transition, frequency, uncertainty, ffreq, fflux, z, verbose=verbose)
    else:
        chi2 = _serial_fft_zfind(transition, frequency, uncertainty, ffreq, fflux, z, verbose=verbose)
    return z, chi2

def fft_per_pixel(transition, frequency, all_flux, z_start=0, dz=0.01, z_end=10, size=3):
    """ Perform the FFT fitting method on all flux values """
    print('Calculating all FFT fit chi-squared values...')
    verbose, parallel, uncertainty = False, False, 1
    with Pool() as pool:
        jobs = [pool.apply_async(fft_zfind, (transition, frequency, flux, uncertainty, z_start, dz, z_end, verbose, parallel)) for flux in all_flux]
        results = [res.get() for res in tqdm(jobs)]
    all_z = [z[np.argmin(chi2)] for z, chi2 in results]
    z = np.reshape(all_z, (size, size))
    return z


"""
Class for finding the redshift of a source in a FITS file. Exports data to csv files and plots figures.
"""

import warnings
from multiprocessing import Pool
from itertools import zip_longest

import numpy as np
from tqdm import tqdm
from astropy.io import fits
from astropy.wcs import WCS
from photutils.aperture import CircularAperture, CircularAnnulus

from zfinder.flux import calc_beam_area, mp_flux_jobs, serial_flux_jobs
from zfinder.template import template_zfind, template_per_pixel, find_lines
from zfinder.fft import fft_zfind, fft, fft_per_pixel
from zfinder.plotter import Plotter
from zfinder.utils import get_eng_exponent, average_zeroes, wcs2pix, generate_square_world_coords, radec2str, gen_random_coords
from zfinder.uncertainty import z_uncert

warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide", category=RuntimeWarning)
warnings.filterwarnings("ignore", message="divide by zero encountered in divide", category=RuntimeWarning)

class zfinder():
    """ 
    zfinder is used to find the redshift of a fits image via two different methods: 
    
    `template` fits gaussian functions to the data to find redshift and
    
    `zfft` performs the fast fourier transform on the flux data to find redshift. 
    
    These methods will create and save a series of plots and csv files with raw data
    by default. Can be changed with the `showfig` and `export` parameters.
    
    If running in a `.py` file, you may need to add the following to your code.
    Otherwise, jupyter notebooks should work fine.
    
    
    >>> if __name__ == '__main__':
    >>>     source = zfinder(fitsfile, ra, dec, aperture_radius, transition)

    Parameters
    ----------
    fitsfile : `.fits`
        A .fits image file
    
    ra : list
        Right ascension of the target [h, m, s]
    
    dec : list
        Declination of the target [d, m, s, esign]
    
    aperture_radius : float
        Radius of the aperture to use over the source (pixels)

    transition : float
        The first transition frequency of the target element or molecule (GHz)
        
    bkg_radius : float, optional
        Radius of the background annulus (pixels). Default=50
    
    beam_tolerance : float, optional
        The tolerance of the beam area (arcsec). Default=1
    
    showfig : bool, optional
        If True, show the figures. Default=True
    
    savefig : bool, optional
        If True, save the figures. Default=True
    
    export : bool, optional
        If True, export the data to csv files. Default=True
        
    Methods
    -------
    get_freq()
        Caclulate the frequency axis
    
    get_flux()
        For every frequency channel, find the flux and associated uncertainty at a position
    
    get_all_flux()
        Get the flux values for all ra and dec coordinates
    
    template()
        Performs the template redshift finding method
    
    template_pp()
        Performs the template redshift finding method in a square around the target ra and dec
    
    fft()
        Performs the fft redshift finding method
    
    fft_pp()
        Performs the fft redshift finding method in a square around the target ra and dec
    
    fft_uncert()
        Calculates the uncertainty on the fft flux. 
    
    Examples
    --------
    >>> from zfinder import zfinder
    >>>
    >>> fitsfile = '0856_cube_c0.4_nat_80MHz_taper3.fits'
    >>> ra = '08:56:14.8'
    >>> dec = '02:24:00.6'
    >>> aperture_radius = 3
    >>> transition = 115.2712
    >>> z_start = 0
    >>> dz = 0.001
    >>> z_end = 10
    >>> 
    >>> source = zfinder(fitsfile, ra, dec, aperture_radius, transition)
    >>> source.template(z_start, dz, z_end)
    >>> source.fft(z_start, dz, z_end)
    >>>
    >>> # Once redshift is found, narrow down the redshift range and run per pixel methods
    >>> aperture_radius_pp = 0.5
    >>> size = 15
    >>> z_start = 5.4
    >>> dz = 0.001
    >>> z_end = 5.7
    >>> source.fft_pp(size, z_start, dz, z_end, aperture_radius_pp)
    >>> source.template_pp(size, z_start, dz, z_end, aperture_radius_pp)
    """
    
    def __init__(self, fitsfile, ra, dec, aperture_radius, transition, bkg_radius=50, beam_tolerance=1, showfig=True, savefig=True, export=True):
        self._fitsfile = fitsfile
        self._hdr = fits.getheader(fitsfile)
        self._data = fits.getdata(fitsfile)[0]
        self._bin_hdu = fits.open(fitsfile)[1]
        self._ra = ra
        self._dec = dec
        self._aperture_radius = aperture_radius
        self._transition = transition
        self._bkg_radius = bkg_radius
        self._beam_tolerance = beam_tolerance
        self._export = export
        self._plotter = Plotter(showfig=showfig, savefig=savefig)

        # Ignore warnings
        warnings.filterwarnings("ignore", module='astropy.wcs.wcs')
        warnings.filterwarnings("ignore", message='Metadata was averaged for keywords CHAN,POL', category=UserWarning)
    
    def get_freq(self):
        """ 
        Caclulate the frequency axis list (x-axis) of the flux
        """
        # Get frequency axis
        start = self._hdr['CRVAL3']
        increment = self._hdr['CDELT3']
        length = self._hdr['NAXIS3']
        end = start + length * increment

        # Create frequency axis
        frequency = np.linspace(start, end, length)

        # Normalise to engineering notation
        self._freq_exponent = get_eng_exponent(frequency[0])
        frequency = frequency / 10**self._freq_exponent
        return frequency
    
    def get_flux(self, verbose=True, parallel=True):
        """ 
        For every frequency channel, find the flux and associated uncertainty at a position

        Paramters
        ---------
        verbose : bool, optional
            If True, print progress. Default=True
        
        parallel : bool, optional
            If True, use multiprocessing. Default=True

        Returns
        -------
        flux : list
            A list of flux values from each frequency channel

        f_uncert : list
            A list of flux uncertainty values for each flux measurement        
        """

        # Calculate area of the beam
        beam_area = calc_beam_area(self._bin_hdu, self._beam_tolerance)
        pix2deg = self._hdr['CDELT2']  # Pixel to degree conversion factor

        # The position to find the flux at
        position = wcs2pix(self._ra, self._dec, self._hdr)

        # Setup the apertures
        inner_radius = 2*self._aperture_radius
        outter_radius = 3*self._aperture_radius
        aperture = CircularAperture(position, self._aperture_radius)
        annulus = CircularAnnulus(position, inner_radius, outter_radius)

        # Process the flux arrays
        if verbose:
            print('Calculating flux values...')
        if parallel:
            flux, flux_uncert = mp_flux_jobs(self._data, aperture, annulus, (self._bkg_radius, self._bkg_radius), pix2deg, beam_area, verbose)
        else:
            flux, flux_uncert = serial_flux_jobs(self._data, aperture, annulus, (self._bkg_radius, self._bkg_radius), pix2deg, beam_area, verbose)

        # Average zeroes so there isn't div by zero error later
        flux_uncert = average_zeroes(flux_uncert)

        # Normalise to engineering notation
        self._flux_exponent = get_eng_exponent(np.max(flux))
        flux = flux / 10**self._flux_exponent
        flux_uncert = flux_uncert / 10**self._flux_exponent
        return flux, flux_uncert

    def get_all_flux(self, all_ra, all_dec, aperture_radius=None):
        """ 
        Get the flux values for all ra and dec coordinates
        
        Paramters
        ---------
        all_ra : list
            A list of all ra coordinates
            
        all_dec : list
            A list of all dec coordinates
        
        aperture_radius : float
            Radius of the aperture to use over the source (pixels). If None, use the aperture_radius
        """
        if aperture_radius is None:
            aperture_radius = self._aperture_radius
        print('Calculating all flux values...')       
        with Pool() as pool:
            jobs = [pool.apply_async(_mp_all_flux, (self._fitsfile, r, d, aperture_radius, \
                self._transition, self._bkg_radius, self._beam_tolerance)) for r, d in zip(all_ra, all_dec)]
            results = [res.get() for res in tqdm(jobs)]
        all_flux, all_uncert = zip(*results)
        return all_flux, all_uncert

    def _calc_freq_flux(self, verbose, parallel):
        """ Calculate the frequency and flux """
        frequency = self.get_freq()
        flux, flux_uncert = self.get_flux(verbose, parallel)
        return frequency, flux, flux_uncert
    
    def _z_uncert(self, z, chi2, sigma):
        """ Caclulate the uncertainty on the best fitting redshift """
        peaks, _, _ = find_lines(self._flux)
        reduced_sigma = sigma**2 / (len(self._flux) - 2*len(peaks) - 1)
        neg, pos = z_uncert(z, chi2, reduced_sigma)
        return neg, pos

    def template(self, z_start=0, dz=0.001, z_end=10, sigma=1, verbose=True, parallel=True):
        """ 
        Perform the template redshift finding method
        
        Parameters
        ----------
        z_start : float, optional
            The starting redshift to search from. Default=0
            
        dz : float, optional
            The step size of the redshift search. Default=0.01
        
        z_end : float, optional
            The ending redshift to search to. Default=10
        
        sigma : float, optional
            The significance level to calculate the redshift uncertainty. Default=1
        
        verbose : bool, optional
            If True, print progress. Default=True
        
        parallel : bool, optional
            If True, use multiprocessing. Default=True
        
        Returns
        -------
        z : float
            Best fit redshift 
        
        z_uncert : tuple(float, float)
            Lower and upper uncertainty on the best fit redshift
        """
        # Calculate the frequency and flux
        if not hasattr(self, '_frequency'):
            self._frequency, self._flux, self._flux_uncert = self._calc_freq_flux(verbose, parallel)

        # Calculate the template chi2
        z, chi2 = template_zfind(self._transition, self._frequency, self._flux, self._flux_uncert, z_start, dz, z_end, verbose, parallel)

        # Plot the template chi2 and flux
        self._plotter.plot_chi2(z, dz, chi2, title='Template')
        self._plotter.plot_template_flux(self._transition, self._frequency, self._freq_exponent, self._flux, self._flux_exponent)
        if self._export:
            self._plotter._export_template_data(self._fitsfile, self._ra, self._dec, self._aperture_radius, self._transition, 
                filename='template.csv', sigma=sigma, flux_uncert=self._flux_uncert)
        neg, pos = self._z_uncert(z, chi2, sigma)
        return z[np.argmin(chi2)], (neg, pos)

    def template_pp(self, size, z_start=0, dz=0.001, z_end=10, aperture_radius_pp=0.5, flux_limit=0.001, contfile=None):
        """ 
        Performs the template redshift finding method in a square around the target ra and dec.
        
        Not recommended for very large redshift search ranges. Remember to narrow down the redshift
        range before using this method. Recommended to have 100-300 redshifts to check per pixel.
        
        z_start = 5.4, dz = 0.001, z_end = 5.7 is a good range for GLEAM J0856.
        
        z_start = 4.28, dz = 0.0001, z_end = 4.31 is a good range for SPT 0345-47.
        
        Parameters
        ----------
        size : int
            The size of a square (centred on ra and dec) to calculate redshifts in (pixels). size=15 is 15x15 pixels
        
        z_start : float, optional
            The starting redshift to search from. Default=0
        
        dz : float, optional
            The step size of the redshift search. Default=0.01
        
        z_end : float, optional
            The ending redshift to search to. Default=10
        
        aperture_radius_pp : float, optional
            Radius of the aperture to use over the source (pixels). Default=0.5
        
        flux_limit : float, optional
            The minimum flux value to calculate redshifts for. Default=0.001
            If contfile is specified, this is the minimum continuum flux.
        
        contfile : str, optional
            The fits file containing the continuum data. Default=None
        """
        # If the other pp method has not been run, calculate all fluxes and uncertainties
        if not hasattr(self, '_all_flux'):
            all_ra, all_dec = generate_square_world_coords(self._fitsfile, self._ra, self._dec, size, aperture_radius_pp)
            self._all_flux, self._flux_uncertainty = self.get_all_flux(all_ra, all_dec, aperture_radius_pp)            

        # Calculate the template chi2 for each pixel
        frequency = self.get_freq()
        z = template_per_pixel(self._transition, frequency, self._all_flux, self._flux_uncertainty, z_start, dz, z_end, size)
        
        # Plot the template pp heatmap
        if self._export:
            header = ['fitsfile', 'ra', 'dec', 'aperture_radius', 'transition', 'size', 'flux_limit', 'contfile']
            data = [[self._fitsfile, self._ra, self._dec, aperture_radius_pp, self._transition, size, flux_limit, contfile], ['']]
            self._plotter._write_method_to_csv('template_per_pixel.csv', header, data)
            self._plotter._export_heatmap_data('template_per_pixel.csv', 'a', z) # export redshifts to csv
        self._plotter.plot_heatmap(ra=self._ra, dec=self._dec, hdr=self._hdr, data=self._data, size=size, \
            z=z, title='Template', aperture_radius=aperture_radius_pp, flux_limit=flux_limit, export=self._export, contfile=contfile) # Plot the template pp heatmap

    def fft(self, z_start=0, dz=0.001, z_end=10, sigma=1, verbose=True, parallel=True, uncertainty=False):
        """ 
        Perform the fft redshift finding method
        
        Parameters
        ----------
        z_start : float, optional
            The starting redshift to search from. Default=0
            
        dz : float, optional
            The step size of the redshift search. Default=0.001
        
        z_end : float, optional
            The ending redshift to search to. Default=10
        
        sigma : float, optional
            The significance level to calculate the redshift uncertainty. Default=1
            
        verbose : bool, optional
            If True, print progress. Default=True
        
        parallel : bool, optional
            If True, use multiprocessing. Default=True
        
        uncertainty : list, bool, optional
            A list of the uncertainty values calculated by `zfinder.fft_uncert`. 
            If True, read the uncertainty values from the csv file.
            If False, no uncertainty values will be used for redshift calculation
        
        Returns
        -------
        z : float
            Best fit redshift
        
        z_uncert : tuple(float, float)
            Lower and upper uncertainty on the best fit redshift
        """
        # Calculate the frequency and flux
        if not hasattr(self, '_frequency'):
            self._frequency, self._flux, self._flux_uncert = self._calc_freq_flux(verbose, parallel)
        
        # Calculate the fft frequency and flux
        self._ffreq, self._fflux = fft(self._frequency, self._flux)
        
        if uncertainty is False:
            uncertainty = 1
        if uncertainty is True:
            uncertainty = self.read_fft_uncert()

        # Calculate the fft chi2
        z, chi2 = fft_zfind(self._transition, self._frequency, self._flux, uncertainty, z_start, dz, z_end, verbose, parallel)

        # Plot the fft chi2 and flux
        self._plotter.plot_chi2(z, dz, chi2, title='FFT')
        self._plotter.plot_fft_flux(self._transition, self._frequency, self._ffreq, self._fflux)
        if self._export:
            self._plotter._export_fft_data(self._fitsfile, self._ra, self._dec, self._aperture_radius, self._transition, 
                filename='fft.csv', sigma=sigma, frequency=self._frequency, flux=self._flux, flux_uncert=uncertainty)
        neg, pos = self._z_uncert(z, chi2, sigma)
        return z[np.argmin(chi2)], (neg, pos)

    def fft_pp(self, size, z_start=0, dz=0.01, z_end=10, aperture_radius_pp=0.5, flux_limit=0.001, contfile=None):
        """ 
        Performs the fft redshift finding method in a square around the target ra and dec
        
        Not recommended for very large redshift search ranges. Remember to narrow down the redshift
        range before using this method. Recommended to have 100-300 redshifts to check per pixel.
        
        z_start = 5.4, dz = 0.001, z_end = 5.7 is a good range for GLEAM J0856.
        
        z_start = 4.28, dz = 0.0001, z_end = 4.31 is a good range for SPT 0345-47.
        
        Parameters
        ----------
        size : int
            The size of a square (centred on ra and dec) to calculate redshifts in (pixels). size=15 is 15x15 pixels
            
        z_start : float, optional
            The starting redshift to search from. Default=0
        
        dz : float, optional
            The step size of the redshift search. Default=0.01
        
        z_end : float, optional
            The ending redshift to search to. Default=10
        
        aperture_radius_pp : float, optional
            Radius of the aperture to use over the source (pixels). Default=0.5
        
        flux_limit : float, optional
            The minimum flux value to calculate redshifts for. Default=0.001
            If contfile is specified, this is the minimum continuum flux.
        
        contfile : str, optional
            The fits file containing the continuum data. Default=None
        """      
        # If the other pp method has not been run, calculate all fluxes and uncertainties
        if not hasattr(self, '_all_flux'):
            all_ra, all_dec = generate_square_world_coords(self._fitsfile, self._ra, self._dec, size, aperture_radius_pp)
            self._all_flux, self._flux_uncertainty = self.get_all_flux(all_ra, all_dec, aperture_radius_pp)
        
        # Calculate the fft chi2 for each pixel
        frequency = self.get_freq()
        z = fft_per_pixel(self._transition, frequency, self._all_flux, z_start, dz, z_end, size)

        # Plot the fft pp heatmap
        if self._export:
            headings = ['fitsfile', 'ra', 'dec', 'aperture_radius', 'transition', 'size', 'flux_limit', 'contfile']
            data = [[self._fitsfile, self._ra, self._dec, aperture_radius_pp, self._transition, size, flux_limit, contfile], ['']]
            self._plotter._write_method_to_csv('fft_per_pixel.csv', headings, data)
            self._plotter._export_heatmap_data('fft_per_pixel.csv', 'a', z) # export redshifts to csv
        self._plotter.plot_heatmap(ra=self._ra, dec=self._dec, hdr=self._hdr, data=self._data, size=size, \
            z=z, title='FFT', aperture_radius=aperture_radius_pp, flux_limit=flux_limit, export=self._export, contfile=contfile) # Plot the template pp heatmap
    
    def fft_uncert(self, n=100, radius=50, min_spread=1):
        """ Doc string here """
        # Find x,y coordinates of the target
        wcs = WCS(self._hdr, naxis=2)
        center_x, center_y = wcs2pix(self._ra, self._dec, self._hdr)
        x, y = gen_random_coords(n, radius, [center_x, center_y], min_spread)
        # self._x_coords, self._y_coords = x, y
            
        # Convert x, y pix coordinates to world ra and dec
        ra, dec = wcs.all_pix2world(x, y, 1)
        all_ra, all_dec = radec2str(ra, dec)

        # Get the flux values for all ra and dec coordinates
        all_flux, _ = self.get_all_flux(all_ra, all_dec, self._aperture_radius)
        freq = self.get_freq()
        
        # Calculate the FFT for each flux
        all_fflux = np.transpose([fft(freq, flux)[1] for flux in all_flux])
        
        # Calculate the standard deviation of all the fflux channels
        all_std = np.std(all_fflux, axis=1)
        
        # column headers
        field_names = ['std', 'x_centre', 'y_centre', 'x', 'y', 'ra', 'dec', 'n', 'radius', 'min_spread', 'fitsfile']
        dtype = [(name, object) for name in field_names]
        data = np.array(list(zip_longest(
            all_std, [center_x], [center_y], x, y, all_ra, all_dec, [n], [radius], 
            [min_spread], [self._fitsfile], fillvalue='')), dtype=dtype)
            
        # Write the data to the csv
        if self._export:
            np.savetxt('.csv', data, delimiter=',', fmt='%s', header=','.join(data.dtype.names))
        self._plotter.plot_coords(center_x, center_y, x, y, radius, fitsfile=self._fitsfile)
        return all_std
    
    @staticmethod
    def read_fft_uncert(filename='fft_uncertainty.csv'):
        """ Read the fft uncertainty csv file """
        std = np.genfromtxt(filename, delimiter=',', usecols=(0)).T
        return std
        
def _mp_all_flux(fitsfile, ra, dec, aperture_radius, transition, bkg_radius, beam_tolerance):
    """ Get the flux values for a single ra and dec coordinate """
    flux, flux_uncert = zfinder(fitsfile, ra, dec, aperture_radius, transition, bkg_radius, beam_tolerance).get_flux(verbose=False, parallel=False)
    return flux, flux_uncert


import pytest

import numpy as np

from zfinder.finder import zfinder
from zfinder.template import template_zfind

def test_template():
    """ Test the template method """

    # Load in the data
    fitsfile = 'zfinder/0856_cube_c0.4_nat_80MHz_taper3.fits'
    ra = '08:56:14.8'
    dec = '02:24:00.6'
    aperture_radius = 3
    transition = 115.2712

    # Calculate the frequency and flux
    source = zfinder(fitsfile, ra, dec, aperture_radius, transition)
    freq = source.get_freq()
    flux, flux_uncert = source.get_flux(verbose=False)
    
    # Calculate the template chi2
    z, chi2 = template_zfind(transition, freq, flux, flux_uncert)

    # Find the best fit redshift
    lowest_index = np.argmin(chi2)
    best_fit_redshift = z[lowest_index]

    assert best_fit_redshift == pytest.approx(5.55)

if __name__ == '__main__':
    test_template()


""" 
Helper functions for calculating fluxes from FITS files
"""

import warnings
from multiprocessing import Pool

import numpy as np
from tqdm import tqdm
from radio_beam import Beam
from photutils.background import Background2D
from photutils.aperture import ApertureStats, aperture_photometry

def calc_beam_area(bin_hdu, tolerance=1):
    """ Caclulate the corrected beam area (from Jy/beam to Jy) """
    beam = Beam.from_fits_bintable(bin_hdu, tolerance)
    bmaj = beam.major.value / 3600  # in degrees
    bmin = beam.minor.value / 3600  # in degrees
    beam_area = 1.1331 * bmaj * bmin
    return beam_area

def mp_flux_jobs(data, aperture, annulus, bkg_radius, pix2deg, beam_area, verbose):
    """ Process the flux arrays using multiprocessing """
    with Pool() as pool:
        jobs = [pool.apply_async(process_channel_data, 
                (channel, aperture, annulus, bkg_radius, pix2deg, beam_area)) 
                for channel in data]
        results = [res.get() for res in tqdm(jobs, disable=not verbose)]
    flux, flux_uncert = zip(*results)
    return np.array(flux), np.array(flux_uncert)

def serial_flux_jobs(data, aperture, annulus, bkg_radius, pix2deg, beam_area, verbose):
    """ Process the flux arrays serially """
    flux, flux_uncert = [], []
    for channel in tqdm(data, disable=not verbose):
        f, u = process_channel_data(channel, aperture, annulus, bkg_radius, pix2deg, beam_area)
        flux.append(f)
        flux_uncert.append(u)
    flux = np.array(flux)
    flux_uncert = np.array(flux_uncert)
    return flux, flux_uncert
    
def process_channel_data(channel, aperture, annulus, bkg_radius, pix2deg, barea):
    """ Function for processing channels in get_flux() """

    # Ignore warnings
    warnings.filterwarnings("ignore", module='photutils.background')

    # Get flux uncertainty
    aperstats = ApertureStats(channel, annulus)
    flux_uncert = aperstats.std

    # Get background flux
    bkg = Background2D(channel, bkg_radius).background

    # Calculate the sum of pixels in the aperture
    apphot = aperture_photometry(channel - bkg, aperture)
    apsum = apphot['aperture_sum'][0]

    # Calculate corrected flux
    flux = apsum*(pix2deg**2)/barea
    return flux, flux_uncert


from numpy import sqrt, argmin
from scipy.optimize import curve_fit

def quadratic(x, a, b, c):
    """ Fit a quadratic to 3 points closest to the minimum chi-squared """
    y = a*(x-b)**2 + c
    return y

def solve_quadratic(y, a, b, c):
    """ Solve the quadratic to find x at a certain sigma """
    neg = -sqrt((y-c)/a) + b
    pos = sqrt((y-c)/a) + b
    return neg, pos

def z_uncert(z, chi2, ssigma):
    """
    Caclulate the uncertainty in the best fitting redshift

    Parameters
    ----------
    z : list
        A list of redshift values
    
    chi2 : list
        A list of chi-squared values

    ssigma : float
        The squared significance level of the uncertainty. ssigma = sigma**2
    
    Returns
    -------
    neg_uncert : float
        The left (- / negative) uncertainty

    pos_uncert : float
        The right (+ / positive) uncertainty
    """
    
    lowest_chi2 = argmin(chi2)
    best_z = z[lowest_chi2]
    min_chi2 = min(chi2)

    # Get points left and right of lowest_y that is above lowest_y + sigma
    left = lowest_chi2-1
    right = lowest_chi2+1

    # Create the axis arrays for the line between the left, lowest, and right point
    line_points_z = [z[left], best_z, z[right]]
    line_points_chi2 = [chi2[left], min_chi2, chi2[right]]
    
    # Find the parameters of the parabola that connects left, lowest, & right
    parabola_params, covars = curve_fit(lambda x, a: quadratic(x, a, b=best_z, c=min_chi2), line_points_z, line_points_chi2)

    # Calculate the z points above lowest_chi2 + sigma 
    neg_z, pos_z = solve_quadratic(min_chi2 + ssigma, *parabola_params, b=best_z, c=min_chi2)

    # Calculate the uncertainty in z
    neg_uncert_z = best_z - neg_z
    pos_uncert_z = pos_z - best_z
    return neg_uncert_z, pos_uncert_z